package ru.dbotthepony.kstarbound.client.render

import org.lwjgl.opengl.GL11.GL_COLOR_BUFFER_BIT
import org.lwjgl.opengl.GL11.GL_RGBA
import org.lwjgl.opengl.GL11.glClear
import ru.dbotthepony.kstarbound.client.gl.BlendFunc
import ru.dbotthepony.kstarbound.client.gl.GLFrameBuffer
import ru.dbotthepony.kstarbound.client.gl.GLStateTracker
import ru.dbotthepony.kstarbound.client.gl.GLTexture2D
import ru.dbotthepony.kstarbound.client.gl.checkForGLError
import ru.dbotthepony.kstarbound.client.gl.vertex.QuadTransformers
import ru.dbotthepony.kstarbound.client.gl.vertex.quad
import ru.dbotthepony.kvector.matrix.Matrix4fStack
import ru.dbotthepony.kvector.vector.Color
import ru.dbotthepony.kvector.vector.nfloat.Vector2f

// https://slembcke.github.io/SuperFastHardShadows
class LightRenderer(val state: GLStateTracker) {
	fun interface ShadowGeometryRenderer {
		fun render(lightPosition: Vector2f, stack: Matrix4fStack)
	}

	private val geometry = ArrayList<ShadowGeometryRenderer>()

	fun addShadowGeometry(geometry: ShadowGeometryRenderer): LightRenderer {
		this.geometry.add(geometry)
		return this
	}

	fun removeShadowGeometry(geometry: ShadowGeometryRenderer): Boolean {
		return this.geometry.remove(geometry)
	}

	/**
	 * Сюда происходит рендер маски света и самого света
	 */
	private val framebufferRender = GLFrameBuffer(state)

	/**
	 * Сюда накапливается отрисованный свет
	 */
	private val framebufferAccumulator = GLFrameBuffer(state)

	val outputTexture: GLTexture2D? get() = framebufferAccumulator.texture

	fun resizeFramebuffer(width: Int, height: Int) {
		framebufferRender.reattachTexture(width, height, GL_RGBA)
		framebufferAccumulator.reattachTexture(width, height, GL_RGBA)
	}

	fun begin() {
		state.ensureSameThread()

		if (!framebufferRender.isComplete || !framebufferAccumulator.isComplete) {
			return
		}

		val old = state.clearColor
		state.clearColor = Color.BLACK

		framebufferRender.bind()
		glClear(GL_COLOR_BUFFER_BIT)
		checkForGLError()
		framebufferAccumulator.bind()
		glClear(GL_COLOR_BUFFER_BIT)
		checkForGLError()
		state.clearColor = old

		framebufferAccumulator.unbind()
	}

	/**
	 * Отрисовывает результат на весь текущий framebuffer в режиме additive
	 */
	fun renderOutputAdditive() {
		val oldFunc = state.blendFunc

		try {
			state.activeTexture = 0
			state.texture2D = outputTexture
			state.blendFunc = BlendFunc.ADDITIVE
			state.programs.textureQuad.run(0)
		} finally {
			state.blendFunc = oldFunc
		}
	}

	/**
	 * Отрисовывает результат на весь текущий framebuffer в режиме additive
	 */
	fun renderOutputMultiplicative() {
		val oldFunc = state.blendFunc

		try {
			state.activeTexture = 0
			state.texture2D = outputTexture
			state.blendFunc = BlendFunc.MULTIPLY_WITH_ALPHA
			state.programs.textureQuad.run(0)
		} finally {
			state.blendFunc = oldFunc
		}
	}

	fun renderLight(
		position: Vector2f,
		color: Color = Color.WHITE,
		radius: Float = 10f,
		stack: Matrix4fStack = state.matrixStack
	) {
		state.ensureSameThread()

		if (!framebufferRender.isComplete || !framebufferAccumulator.isComplete) {
			return
		}

		val oldFunc = state.blendFunc

		// отрисовка
		try {
			framebufferRender.bind()

			// state.programs.colorQuad.clearAlpha()

			// Геометрия
			val old = state.clearColor
			state.clearColor = Color.BLACK
			glClear(GL_COLOR_BUFFER_BIT)
			checkForGLError()
			state.clearColor = old

			state.programs.lightOccluder.use()
			state.programs.lightOccluder.transform.set(stack.last)
			state.programs.lightOccluder.lightPosition.set(position)

			state.blendFunc = BlendFunc.ONLY_ALPHA

			for (renderer in geometry) {
				renderer.render(position, stack)
			}

			state.programs.light.use()
			state.programs.light.transform.set(stack.last)
			state.programs.light.baselineColor.set(color)

			// Свет
			val builder = state.programs.light.builder

			builder.begin()
			builder.quad(position.x - radius, position.y - radius, position.x + radius, position.y + radius, QuadTransformers.uv())
			builder.upload()

			state.blendFunc = BLEND_MODE
			builder.draw()
		} finally {
			state.blendFunc = oldFunc
			framebufferRender.unbind()
		}

		// накопление
		try {
			framebufferAccumulator.bind()

			state.blendFunc = BlendFunc.ADDITIVE
			state.activeTexture = 0
			state.texture2D = framebufferRender.texture
			state.programs.textureQuad.run(0)
		} finally {
			state.blendFunc = oldFunc
			framebufferAccumulator.unbind()
		}
	}

	companion object {
		val BLEND_MODE = BlendFunc(
			BlendFunc.Func.DST_ALPHA,
			BlendFunc.Func.ZERO,
			BlendFunc.Func.ZERO,
			BlendFunc.Func.ONE,
		)
	}
}
